

		
(define-syntax dolist
  (syntax-rules
   ()
   ((dolist (var l) body1 ...)
    (let loop-label ((itt-var l))
      (if (null? itt-var)
	  (void)
	  (let ((var (car itt-var)))
	    (begin
	      (begin body1 ...)
	      (loop-label (cdr itt-var)))))))))



(define-syntax dotimes
  (syntax-rules
   ()
   ((dotimes (var limit) body1 ...)
    (let ((v-limit limit))
      (let loop-label ((var 0))
	(if (= var v-limit)
	    (void)
	    (begin
	      (begin body1 ...)
	      (loop-label (+ 1 var)))))))))



(define (pormat-function control-string . values)

  (let loop ((i 0)(values values))
    (when (not (= i (string-length control-string)))
      (let ((c (string-ref control-string i)))

	(if (eqv? c #\~)

	    (let ((c2 (string-ref control-string (+ 1 i))))
	      (cond ((eqv? c2 #\a)
		     (display (car values))
		     (loop (+ 2 i) (cdr values)))
		    ((eqv? c2 #\%)
		     (newline)
		     (loop (+ 2 i) values))
		    (else
		     (display c)
		     (loop (+ 1 i) values))))

	    (begin
	      (display c)
	      (loop (+ 1 i) values)))))))







(define (revbuilder-new)
  (box '()))


(define (revbuilder-add rb new)
  (set-box! rb (cons new (unbox rb))))


(define (revbuilder-done rb)
  (reverse (unbox rb)))







;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (second x) (cadr x))
(define (third  x) (caddr x))
(define (fourth x) (cadddr x))
    


(define (walk-list l k)
  (let loop ((l l))
    (when (not (null? l))
      (k (car l))
      (loop (cdr l)))))


(define (do-times-up n k) ; equivalent to (for-up 0 (- n 1) k)
  (let loop ((i 0))
    (when (< i n)
      (k i)
      (loop (+ i 1)))))


(define (read-file name)
  (let ((the-file (open-input-file name)))
    (let loop ((try (read the-file))(accum '()))
      (if (eof-object? try)
	  (begin
	    (close-input-port the-file)
	    (reverse accum))
	  (loop (read the-file)(cons try accum))))))



(define (read-int file)
  (let ((val (read file)))
    (if (integer? val)
	val
	(error "bogus value"))))


(define (read-cform-trainset name)
  (let* ((the-file (open-input-file name))
	 (n-strings (read-int the-file))
	 (new-alphabet-size (read-int the-file))
	 (rb (revbuilder-new)))
    (set-box! alphabet-size new-alphabet-size)
    (do-times-up n-strings
      (lambda (i)
	(let* ((label (read-int the-file))
	       (string-len (read-int the-file))
	       (rb2 (revbuilder-new)))
	  (do-times-up string-len
	    (lambda (j)
	      (revbuilder-add rb2 (read-int the-file))))
	  (revbuilder-add rb (cons label (revbuilder-done rb2))))))
    (revbuilder-done rb)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (make-queue)
  (cons '() '()))

(define (queue-empty? q)
  (and (null? (car q))
       (null? (cdr q))))

(define (queue-push q obj)
  (set-car! q (cons obj (car q))))

(define (queue-pop q)
  (normalize-queue! q)
  (let ((head (car (cdr q))))
    (set-cdr! q (cdr (cdr q)))
    head))

(define (normalize-queue! q)
  (if (null? (cdr q))
      (begin (set-cdr! q (reverse (car q)))
	     (set-car! q '()))))

(define (queue-head q)
  (normalize-queue! q)
  (car (cdr q)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (make-counter)
  (let ((count (box 0)))
    (lambda ()
      (let ((old-count (unbox count)))
	(set-box! count (+ 1 (unbox count)))
      old-count))))
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(define (make-table) (vector 'table 0 1 (make-vector 1 '())))

; (define (table-get-count    table) (vector-ref table 1))
; (define (table-get-n-alists table) (vector-ref table 2))
; (define (table-get-alists   table) (vector-ref table 3))
; 
; (define (table-set-count    table new) (vector-set! table 1 new))
; (define (table-set-n-alists table new) (vector-set! table 2 new))
; (define (table-set-alists   table new) (vector-set! table 3 new))



(define-syntax table-get-count
  (syntax-rules ()
   ((table-get-count table) (vector-ref table 1))))

(define-syntax table-set-count
  (syntax-rules ()
   ((table-set-count table new) (vector-set! table 1 new))))


(define-syntax table-get-n-alists
  (syntax-rules ()
   ((table-get-n-alists table) (vector-ref table 2))))

(define-syntax table-set-n-alists
  (syntax-rules ()
   ((table-set-n-alists table new) (vector-set! table 2 new))))


(define-syntax table-get-alists
  (syntax-rules ()
   ((table-get-alists table) (vector-ref table 3))))

(define-syntax table-set-alists
  (syntax-rules ()
   ((table-set-alists table new) (vector-set! table 3 new))))


(define (table-test table key)
  (let* ((n-alists (table-get-n-alists table))
	 (alists   (table-get-alists table))
	 (hash (modulo key n-alists))
	 (probe (assq key (vector-ref alists hash))))
    (if probe #t #f)))

(define (table-lookup table key)
  (let* ((n-alists (table-get-n-alists table))
	 (alists   (table-get-alists table))
	 (hash (modulo key n-alists))
	 (probe (assq key (vector-ref alists hash))))
    (if probe
	(cdr probe)
	(error "table-lookup failed"))))


(define (table-expand table)
  (let* ((old-n-alists (table-get-n-alists table))
	 (old-alists   (table-get-alists table))
	 (new-n-alists (* 10 old-n-alists))
	 (new-alists   (make-vector new-n-alists '())))
    (dotimes (i old-n-alists)
      (dolist (a (vector-ref old-alists i))
	(let ((new-hash (modulo (car a) new-n-alists)))
	  (vector-set! new-alists new-hash
		       (cons a (vector-ref new-alists new-hash))))))
    (table-set-n-alists table new-n-alists)
    (table-set-alists   table new-alists)))
    

(define (table-set! table key value)
  (let* ((n-alists (table-get-n-alists table))
	 (alists   (table-get-alists table))
	 (hash (modulo key n-alists))
	 (probe (assq key (vector-ref alists hash))))
    (if probe
	(set-cdr! probe value)
	(begin
	  (vector-set! alists hash
		       (cons (cons key value) (vector-ref alists hash)))
	  (table-set-count table (1+ (table-get-count table)))
	  (when (> (table-get-count table) (* 10 n-alists))
	    (table-expand table))))))


(define (table-count table)
  (table-get-count table))


(define (walk-table k table)
  (let ((n-alists (table-get-n-alists table))
	(alists   (table-get-alists table)))
    (dotimes (i n-alists)
      (dolist (p (vector-ref alists i))
	(k (car p)(cdr p))))))


(define (table-test-and-set! table key)
  (let ((probe (table-test table key)))
    (when (not probe)
      (table-set! table key #t))
    probe))







;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;





(define (walk-vector fun vec)
  (let ((veclen (vector-length vec)))
    (let next ((i 0))
      (when (< i veclen)
	(fun (vector-ref vec i))
	(next (1+ i))))))
	 

(define (walk-vector-with-offset offset fun vec)
  (let ((veclen (vector-length vec)))
    (let next ((i offset))
      (when (< i veclen)
	(fun (vector-ref vec i))
	(next (1+ i))))))
	 


(define (vector-replace to from len)
  (do-times-up len
    (lambda (i)
      (vector-set! to i (vector-ref from i)))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




(define alphabet-size (box 2))



; note: for now, using integers for state labels
; -1 unknown
;  0 reject
;  1 accept


; (ml-type-def possible_node
; 	     (nondef)
; 	     (tree-node * * * *))


(define (nondef) 'ndf)
(define (nondef? x) (eq? 'ndf x))




; 
; (define (gloid node)
;     (vector-ref node 0))
; 
; (define (state-label node)
;     (vector-ref node 1))
; 
; (define (father node)
;     (vector-ref node 2))
; 
; (define (get-child node i)
;     (vector-ref node (+ 3 i)))
; 
; 
; (define (set-state-label node val)
;     (vector-set! node 1 val))
; 
; (define (set-father node val)
;     (vector-set! node 2 val))
; 
; (define (set-child node i val)
;     (vector-set! node (+ 3 i) val))
; 
; 
; (define (walk-children fun node)
;     (walk-vector-with-offset 3 fun node))






 (define-syntax gloid
   (syntax-rules ()
    ((gloid node)
     (vector-ref node 0))))
 
 (define-syntax state-label
   (syntax-rules ()
    ((state-label node)
     (vector-ref node 1))))
 
 (define-syntax father
   (syntax-rules ()
    ((father node)
     (vector-ref node 2))))
 
 (define-syntax get-child
   (syntax-rules ()
    ((get-child node i)
     (vector-ref node (+ 3 i)))))
 
 
 (define-syntax set-state-label
   (syntax-rules ()
    ((set-state-label node val)
     (vector-set! node 1 val))))
 
 (define-syntax set-father
   (syntax-rules ()
    ((set-father node val)
     (vector-set! node 2 val))))
 
 (define-syntax set-child
   (syntax-rules ()
    ((set-child node i val)
     (vector-set! node (+ 3 i) val))))
 
 (define-syntax walk-children
   (syntax-rules ()
    ((walk-children fun node)
     (walk-vector-with-offset 3 fun node))))
 



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(define global-node-counter (box 0))

(define (make-node father)
  (let ((old-counter-val (unbox global-node-counter))
	(node (make-vector (+ 3 (unbox alphabet-size)))))
    (set-box! global-node-counter (+ (unbox global-node-counter) 1))
    (vector-set!     node 0 old-counter-val)
    (set-state-label node -1)
    (set-father      node father)
    (do-times-up (unbox alphabet-size)
      (lambda (i)
	(set-child node i (nondef))))
    node))


(define (make-duplicate-node x)
  (copy-vector x))

(define (replace-node-contents to from)
  (vector-replace to from (+ 3 (unbox alphabet-size))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(define (augment-tree curnode string label)
  (if (null? string)
      (set-state-label curnode label)
      (let ((cursym (car string)))
	(when (nondef? (get-child curnode cursym))
	  (set-child curnode cursym (make-node curnode)))
	(augment-tree (get-child curnode cursym)
		      (cdr string) label))))


(define (new-build-tree example-list)
  (let ((root (make-node (nondef))))
    (walk-list example-list
      (lambda (example)
	(augment-tree root (cdr example)(car example))))
    root))


(define (build-tree-from-cform-file name)
  (let* ((the-file (open-input-file name))
	 (n-strings (read-int the-file))
	 (ignore
	  (begin
	    (set-box! alphabet-size (read-int the-file))
	    #f))
	 (root (make-node (nondef))))

    (do-times-up n-strings
      (lambda (i)
	(let* ((label (read-int the-file))
	       (string-len (read-int the-file))
	       (rb2 (revbuilder-new)))
	  (do-times-up string-len
	    (lambda (j)
	      (revbuilder-add rb2 (read-int the-file))))
	  (augment-tree root (revbuilder-done rb2) label))))
    root))
			





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




(define (find-incoming-token node father)
  (let next ((i 0)(n-matches 0)(matching-i -1))
    (if (= i (unbox alphabet-size))
	(if (= n-matches 1)
	    matching-i
	    (error "horrible bug in find-incoming-token"))
	(next (1+ i)
	      (if (eq? (get-child father i) node)
		  (1+ n-matches)
		  n-matches)
	      (if (eq? (get-child father i) node)
		  i
		  matching-i)))))
	    
	    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




(define (fancy-print-tree start-node)
  (let ((seen (make-table))
	(id-table (make-table))
	(todo (make-queue))
	(count (make-counter)))

    (queue-push todo start-node)
    (let next-node ()
      (let ((curnode (queue-pop todo)))
	(when (not (table-test id-table (gloid curnode)))
	  (table-set! id-table (gloid curnode) (count))
	  (walk-children
	   (lambda (x)(when (not (nondef? x)) (queue-push todo x)))
	   curnode)
	  )
	(when (not (queue-empty? todo)) (next-node))))


    (pormat-function "~a ~a~%" (count) (unbox alphabet-size))

    (queue-push todo start-node)
    (let next ()
      (let ((curnode (queue-pop todo)))
	(when (not (table-test-and-set! seen (gloid curnode)))

	  (pormat-function "~a ~a "
			   (table-lookup id-table (gloid curnode))
			   (state-label curnode))

	  (walk-children
	   (lambda (x) (if (nondef? x)
			   (pormat-function "? ")
			   (begin
			     (queue-push todo x)
			     (pormat-function "~a "
					      (table-lookup id-table
							    (gloid x))))))
	   curnode)
	  (pormat-function "~%")))
      (when (not (queue-empty? todo))(next)))))








;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


; This procedure destructively copies labels and transitions from
; the tree into the graph while checking for labeling conflicts.
; It makes backup copies of modified nodes so that these side effects 
; can be undone if the merge fails.

(define (merge-tree-into-graph tr-node graph-node
			       bailout backup-table splices)

  (when (not (= -1 (state-label tr-node))) ; -1 means unknown.
    (cond ((not (= -1 (state-label graph-node)))
	   (when (not (eq? (state-label graph-node)
			   (state-label tr-node)))
	     (begin (bailout #f)))) ; label conflict, merge fails.
	  (else
	   (backup graph-node backup-table) ; copy tree label into graph
	   (set-state-label graph-node (state-label tr-node)))))

  (do-times-up (unbox alphabet-size)
    (lambda (i)
      (let ((graph-child (get-child graph-node i))
	    (tree-child  (get-child tr-node i)))

	(when (not (nondef? tree-child))
	  (if (not (nondef? graph-child))

              ; continue walking 
	      (merge-tree-into-graph tree-child graph-child
				     bailout backup-table splices)

	      (begin ; splice tree branch into graph
		(backup graph-node backup-table)
		(backup tree-child backup-table)
		(set-child graph-node i tree-child)
		(set-father tree-child graph-node)
		(set-box! splices
			  (cons (cons graph-node tree-child)
				(unbox splices)))
		)))))))





(define (try-merging curnode potential-match backup-table splices)
  (call/cc
   (lambda (merge-failure-bailout)
     (merge-tree-into-graph curnode potential-match
			    merge-failure-bailout
			    backup-table splices)
     #t)))




(define (accept-the-node curnode todo unique-table unique-nodes)
  (set-box! unique-nodes (append (unbox unique-nodes) (list curnode)))
  (table-set! unique-table (gloid curnode) #t)
  (walk-children
   (lambda (x)(when (not (nondef? x)) (queue-push todo x)))
   curnode))

  


; this outer loop visits the tree's nodes in breadth first order.

(define (consider-node todo unique-table unique-nodes)

  (let* ((curnode (queue-pop todo))
	 (curdad  (father curnode))
	 (inchar  (if (nondef? curdad) -1 ; happens only at root
		      (find-incoming-token curnode curdad))))

    ; this inner loop compares nodes with previously accepted states.

    (let find-match ((potential-matches (unbox unique-nodes)))

      (if (null? potential-matches)

          ; no state is compatible, so we view the node as a new state
	  (accept-the-node curnode todo unique-table unique-nodes)

	  (let ((potential-match (car potential-matches))
		(backup-table (make-table))
		(splices (box '())))

	    (set-child curdad inchar potential-match)

	    (if (try-merging curnode potential-match backup-table splices)

		; merge attempt suceeded.  remind ourselves to 
		; visit any new children of previously accepted states.
		(walk-list (unbox splices)
		  (lambda (x)
		    (when (table-test unique-table (gloid (car x)))
		      (queue-push todo (cdr x)))))

		; merge attempt failed.  must undo all side effects.
		(begin
		  (revert-from-backups backup-table)
		  (set-child curdad inchar curnode)
		  (find-match (cdr potential-matches)))) ; end of inner loop
	    )))
		    
    (when (not (queue-empty? todo))
      (consider-node todo unique-table unique-nodes)))) ; end of outer loop





(define (greedily-collapse root)

  (let ((todo         (make-queue))
	(unique-table (make-table))
	(unique-nodes (box '())))

    (queue-push todo root)
    (consider-node todo unique-table unique-nodes)
    root))






;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (backup guy backup-table)
  (when (not (table-test backup-table (gloid guy)))
    (let ((backup (make-duplicate-node guy)))
      (table-set! backup-table (gloid guy) (cons guy backup))
      )))
     

(define (revert-from-backups backup-table)
  (when (not (zero? (table-count backup-table)))
    (walk-table (lambda (id guy-dot-backup)
		  (let ((guy    (car guy-dot-backup))
			(backup (cdr guy-dot-backup)))
		    (replace-node-contents guy backup)))
		backup-table)
    ))

 




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;





(define (main file-name)
  (fancy-print-tree
   (greedily-collapse
    (build-tree-from-cform-file file-name))))



